/**********************************************************************
 Freeciv - Copyright (C) 2003 - The Freeciv Team
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
***********************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "city.h"
#include "combat.h"
#include "game.h"
#include "map.h"
#include "log.h"
#include "pf_tools.h"
#include "player.h"
#include "unit.h"

#include "citytools.h"
#include "settlers.h"

#include "ailog.h"
#include "aitools.h"
#include "aiunit.h"

#include "aiseahunt.h"

/**************************************************************************
  We don't need a hunter in this city if we already have one or we are
  landlocked.
**************************************************************************/
static bool ai_hunter_wanted(struct player *pplayer, struct city *pcity)
{
  if (!is_at_coast(pcity->x, pcity->y)) {
    return FALSE;
  }

  unit_list_iterate(pcity->units_supported, punit) {
    if (ai_hunter_qualify(pplayer, punit)) {
      return FALSE;
    }
  } unit_list_iterate_end;
  unit_list_iterate(map_get_tile(pcity->x, pcity->y)->units, punit) {
    if (ai_hunter_qualify(pplayer, punit)) {
      return FALSE;
    }
  } unit_list_iterate_end;

  return TRUE;
}

/**************************************************************************
  Guess best hunter unit type.
**************************************************************************/
static Unit_Type_id ai_hunter_guess_best(struct city *pcity)
{
  Unit_Type_id bestid = -1;
  int best = 0;

  unit_type_iterate(i) {
    struct unit_type *ut = get_unit_type(i);
    int desire;

    if (ut->move_type != SEA_MOVING
        || !can_build_unit(pcity, i)
        || ut->attack_strength < ut->transport_capacity) {
      continue;
    }

    desire = (ut->hp
              * ut->attack_strength
              * ut->firepower
              * ut->move_rate
              + ut->defense_strength) / TRADE_WEIGHTING; /* WAG */

    if (unit_type_flag(i, F_CARRIER)
        || unit_type_flag(i, F_MISSILE_CARRIER)) {
      desire += desire / 6;
    }
    if (unit_type_flag(i, F_PARTIAL_INVIS)) {
      desire += desire / 4;
    }
    if (unit_type_flag(i, F_NO_LAND_ATTACK)) {
      desire -= desire / 4; /* less flexibility */
    }

    amortize(desire, ut->build_cost / pcity->shield_surplus);

    if (desire > best) {
        best = desire;
        bestid = i;
    }
  } unit_type_iterate_end;

  return bestid;
}

/**************************************************************************
  Check if we want to build a missile for our hunter.
**************************************************************************/
static int ai_hunter_missile_want(struct player *pplayer,
                                  struct city *pcity,
                                  Unit_Type_id *best_unit_type)
{
  int best = -1;
  bool have_hunter = FALSE;

  *best_unit_type = -1;

  unit_list_iterate(map_get_tile(pcity->x, pcity->y)->units, punit) {
    if (ai_hunter_qualify(pplayer, punit)
        && (unit_flag(punit, F_MISSILE_CARRIER)
            || unit_flag(punit, F_CARRIER))) {
      /* There is a potential hunter in our city which we can equip 
       * with a missile. Do it. */
      have_hunter = TRUE;
      break;
    }
  } unit_list_iterate_end;

  if (!have_hunter) {
    return -1;
  }

  unit_type_iterate(i) {
    struct unit_type *ut = get_unit_type(i);
    int desire;

    if (!BV_ISSET(ut->flags, F_MISSILE) || !can_build_unit(pcity, i)) {
      continue;
    }

    desire = (ut->hp
              * ut->attack_strength
              * ut->firepower
              * ut->move_rate) / TRADE_WEIGHTING; /* WAG */

    amortize(desire, ut->build_cost / pcity->shield_surplus);

    if (desire > best) {
        best = desire;
        *best_unit_type = i;
    }
  } unit_type_iterate_end;

  return best;
}

/**************************************************************************
  Check if we want to build a hunter.
**************************************************************************/
void ai_hunter_choice(struct player *pplayer, struct city *pcity,
                      struct ai_choice *choice)
{
  int best_unit_type = ai_hunter_guess_best(pcity);
  int want;
  struct unit *virtualunit;
  bool veteran = player_knows_improvement_tech(pplayer, B_PORT);

  if (best_unit_type == -1 || is_barbarian(pplayer) || !pplayer->is_alive
      || ai_handicap(pplayer, H_TARGETS)
      || !ai_hunter_wanted(pplayer, pcity)) {
    return; /* None available */
  }

  /* Want for a hunter */

  virtualunit = create_unit_virtual(pplayer, pcity, best_unit_type, veteran);
  want = ai_hunter_findjob(pplayer, virtualunit);
  destroy_unit_virtual(virtualunit);
  if (want > choice->want) {
    CITY_LOG(LOGLEVEL_HUNT, pcity, "pri hunter w/ want %d", want);
    choice->choice = best_unit_type;
    choice->want = want;
    choice->type = CT_ATTACKER;
    return;
  }

  /* Want missiles to go with those hunters? */

  want = ai_hunter_missile_want(pplayer, pcity, &best_unit_type);
  if (want > choice->want) {
    CITY_LOG(LOGLEVEL_HUNT, pcity, "pri missile w/ want %d", want);
    choice->choice = best_unit_type;
    choice->want = want;
    choice->type = CT_ATTACKER;
  } else if (want != -1) {
    CITY_LOG(LOGLEVEL_HUNT, pcity, "not pri missile w/ want %d"
             "(old want %d)", want, choice->want);
  }
}

/**************************************************************************
  Does this unit qualify as a hunter? FIXME: Should also check for
  combat damage? Or should repair code preempt this code? Just saying
  FALSE for damaged units does NOT work.
**************************************************************************/
bool ai_hunter_qualify(struct player *pplayer, struct unit *punit)
{
  struct unit_type *punittype = get_unit_type(punit->type);

  if (is_barbarian(pplayer)
      || !is_sailing_unit(punit)
      || punittype->move_rate < 3 * SINGLE_MOVE
      || ATTACK_POWER(punit) <= 1
      || punit->owner != pplayer->player_no) {
    return FALSE;
  }
  if (unit_flag(punit, F_PARTIAL_INVIS)) {
    return TRUE;
  }
  /* TODO: insert better algo here */
  return IS_ATTACKER(punit);
}

/**************************************************************************
  Return want for making this (possibly virtual) unit into a hunter. Sets
  punit->ai.target to target's id.
**************************************************************************/
int ai_hunter_findjob(struct player *pplayer, struct unit *punit)
{
  int best_id = -1, best_val = -1;
  int loglevel = (punit->id == 0) ? LOG_DEBUG : LOGLEVEL_HUNT;

  assert(!is_barbarian(pplayer));
  assert(pplayer->is_alive);
  assert(is_sailing_unit(punit));

  players_iterate(aplayer) {
    if (!aplayer->is_alive || !pplayers_at_war(aplayer, pplayer)) {
      continue;
    }
    unit_list_iterate(aplayer->units, target) {
      struct tile *ptile = map_get_tile(target->x, target->y);
      int dist1, dist2, stackthreat = 0, stackcost = 0;
      struct unit *defender;

      if (ptile->city
          || target->ai.hunted & (1 << pplayer->player_no)
          || !is_ocean(ptile->terrain)
          || !is_sailing_unit(target)
          || !goto_is_sane(punit, target->x, target->y, TRUE)) {
        /* Can't hunt this one. */
        continue;
      }
      if (target->ai.cur_pos && target->ai.prev_pos) {
        dist1 = real_map_distance(punit->x, punit->y, target->ai.cur_pos->x, 
                                  target->ai.cur_pos->y);
        dist2 = real_map_distance(punit->x, punit->y, target->ai.prev_pos->x, 
                                  target->ai.prev_pos->y);
      } else {
        dist1 = dist2 = 0;
      }
      UNIT_LOG(LOG_DEBUG, punit, "considering chasing %s(%d, %d) id %d "
               "dist1 %d dist2 %d", unit_type(target)->name, target->x, 
               target->y, target->id, dist1, dist2);
      /* We can't attack units stationary in cities. */
      if (map_get_city(target->x, target->y) 
          && (dist2 == 0 || dist1 == dist2)) {
        continue;
      }
      /* We can't chase if we aren't faster or on intercept vector */
      if (unit_type(punit)->move_rate < unit_type(target)->move_rate
          && dist1 >= dist2) {
        UNIT_LOG(LOGLEVEL_HUNT, punit, "can't get to %s(%d, %d) was (%d, %d)",
                 unit_type(target)->name, target->x, target->y,
                 target->ai.prev_pos->x, target->ai.prev_pos->y);
        continue;
      }
      unit_list_iterate(ptile->units, sucker) {
        stackthreat += ATTACK_POWER(sucker);
        if (unit_flag(sucker, F_DIPLOMAT)) {
          stackthreat += 500;
        }
        stackcost += unit_type(sucker)->build_cost;
      } unit_list_iterate_end;
      defender = get_defender(punit, target->x, target->y);
      if (stackcost < unit_type(punit)->build_cost
          && unit_win_chance(punit, defender) < 0.6) {
        UNIT_LOG(LOGLEVEL_HUNT, punit, "chickening out from attacking %s"
                 "(%d, %d)", unit_type(defender)->name, defender->x, 
                 defender->y);
        continue;
      }
      stackthreat *= 9; /* WAG */
      stackthreat /= real_map_distance(punit->x, punit->y, 
                                       target->x, target->y) + 1;
      UNIT_LOG(LOG_DEBUG, punit, "considering hunting %s's %s(%d, %d) id "
               "id %d with want %d, dist1 %d, dist2 %d", 
               unit_owner(defender)->name, unit_type(defender)->name, 
               defender->x, defender->y, defender->id, stackthreat, dist1, 
               dist2);
      /* Ok, now we FINALLY have a candidate */
      if (stackthreat > best_val) {
        best_val = stackthreat;
        best_id = target->id;
      }
    } unit_list_iterate_end;
  } players_iterate_end;

  punit->ai.target = best_id;
  if (best_val < MORT) {
    /* Safety against silly missions. Unset our role. */
    best_val = -1;
  } else {
    UNIT_LOG(loglevel, punit, "suggest chasing %d with want %d",
             best_id, best_val);
  }
  return best_val;
}

/**************************************************************************
  Try to shoot our target with a missile. Also shoot down anything that
  might attempt to intercept _us_. We assign missiles to a hunter in
  ai_unit_new_role().
**************************************************************************/
static void ai_hunter_try_launch(struct player *pplayer,
                                 struct unit *punit,
                                 struct unit *target)
{
  int target_sanity = target->id;

  unit_list_iterate(map_get_tile(punit->x, punit->y)->units, missile) {
    struct unit *sucker = NULL;

    if (missile->owner == pplayer->player_no
        && unit_flag(missile, F_MISSILE)) {
      UNIT_LOG(LOGLEVEL_HUNT, missile, "checking for hunt targets");
      simple_unit_path_iterator(missile, pos) {
        if (pos.total_MC > missile->moves_left / SINGLE_MOVE) {
          break;
        }
        unit_list_iterate(map_get_tile(pos.x, pos.y)->units, victim) {
          struct unit_type *ut = unit_type(victim);

          if (!pplayers_at_war(unit_owner(victim), pplayer)) {
            continue;
          }
          if (victim == target) {
            sucker = victim;
            UNIT_LOG(LOGLEVEL_HUNT, missile, "found primary target %d(%d, %d)"
                     " dist %d", victim->id, victim->x, victim->y, 
                     pos.total_MC);
            break; /* Our target! Get him!!! */
          }
          if (ut->move_rate + victim->moves_left > pos.total_MC
              && ATTACK_POWER(victim) > DEFENCE_POWER(punit)
              && (ut->move_type == SEA_MOVING
                  || ut->move_type == AIR_MOVING)) {
            /* Threat to our carrier. Kill it. */
            sucker = victim;
            UNIT_LOG(LOGLEVEL_HUNT, missile, "found aux target %d(%d, %d)",
                     victim->id, victim->x, victim->y);
            break;
          }
        } unit_list_iterate_end;
        if (sucker) {
          break; /* found something - kill it! */
        }
      } simple_unit_path_iterator_end;
      if (sucker) {
UNIT_LOG(LOGLEVEL_HUNT, missile, "attacking!");
        ai_unit_goto(missile, sucker->x, sucker->y);
        sucker = find_unit_by_id(target_sanity); /* Sanity */
        if (sucker && is_tiles_adjacent(sucker->x, sucker->y, missile->x, missile->y)) {
          ai_unit_attack(missile, sucker->x, sucker->y);
        }
        target = find_unit_by_id(target_sanity); /* Sanity */
        break; /* try next missile, if any */
      }
    } /* if */
  } unit_list_iterate_end;
}

/**************************************************************************
  Manage a hunter unit. We assume it has AIUNIT_HUNTER role and a valid
  target in punit->ai.hunt.

  Returns FALSE if we could not use unit. If we return TRUE, unit might
  be dead.
**************************************************************************/
bool ai_hunter_manage(struct player *pplayer, struct unit *punit)
{
  struct unit *target = find_unit_by_id(punit->ai.target);
  int sanity_own = punit->id;
  int sanity_target;

  logdebug_suppress_warning; /* had to put it somewhere */

  CHECK_UNIT(punit);
  assert(punit->ai.ai_role == AIUNIT_HUNTER);

  /* Check that target is valid. */
  if (!target
      || !goto_is_sane(punit, target->x, target->y, TRUE)
      || !pplayers_at_war(pplayer, unit_owner(target))) {
    UNIT_LOG(LOGLEVEL_HUNT, punit, "target vanished");
    ai_unit_new_role(punit, AIUNIT_NONE, -1, -1);
    return FALSE;
  }
  UNIT_LOG(LOGLEVEL_HUNT, punit, "hunting %d(%d, %d)", target->id, target->x,
           target->y);
  sanity_target = target->id;

  /* Check if we can nuke it */
  ai_hunter_try_launch(pplayer, punit, target);

  /* Go towards it. */
  if (!ai_unit_goto(punit, target->x, target->y)) {
    return TRUE;
  }

  /* Check if we can nuke it now */
  ai_hunter_try_launch(pplayer, punit, target);

  /* If we are adjacent - RAMMING SPEED! */
  if (is_tiles_adjacent(punit->x, punit->y, target->x, target->y)) {
    ai_unit_attack(punit, target->x, target->y);
  }

  if (!find_unit_by_id(sanity_own)) {
    return TRUE;
  }
  if (!find_unit_by_id(sanity_target)) {
    UNIT_LOG(LOGLEVEL_HUNT, punit, "mission accomplished");
    ai_unit_new_role(punit, AIUNIT_NONE, -1, -1);
  }

  return TRUE;
}
